
/*
OilWatcherus100 sketch by Steve M. Potter   steve.potter at gmail.com
To monitor heating oil tank level remotely, and send Pushbullet alerts when fuel is getting low.
Uses the Particle Photon board, Adafruit's Verter voltage buck-boost converter, an additional adjustable voltage regulator, and PushBullet
Verter is available in EU from Pimoroni https://shop.pimoroni.com/products/verter-5v-usb-buck-boost-500ma-from-3v-5v-1000ma-from-5v-12v
Uses US-100 ultrasonic rangefinder.

Inspired by code and design by Ventz: https://blog.vpetkov.net/2017/11/12/diy-monitor-heating-oil-tank-gallons-with-pushbullet-sms-and-email-alerting/
and also from his Instructable: https://www.instructables.com/id/Monitor-Heating-Oil-Tank-Gallons-With-Email-SMS-an/

Last update 2018-10-03
Code is kept on http://build.particle.io  called "OilWatcherUS100"

*/
// Libraries to include:
#include "RunningMedian.h"  // by  Rob.Tillaart at gmail.com
#include <ThingSpeak.h>  // for data logging and analysis, e.g. see https://thingspeak.com/channels/377827
#include <OneWire.h>   //  For the DS18B20 digital temperature sensors
#include <DS18B20.h> // For the DS18B20 digital temperature sensors
//#include <NewPing.h> // Improved communication to the HC-SR04 ultrasonic rangefinder, by Tim Eckel
                       // Only for use in analog pulse mode.

// CONSTANTS:
#define VBATT        A1  // blue wire. Monitors the voltage of the battery, through a 1/10 voltage divider.
#define POSCURPIN    A4  // white Wire attached to one side of a 1-ohm shunt resistor in series with the battery, then a voltage divider.
#define NEGCURPIN    A5  // blue Wire attached to the other side of a 1-ohm shunt resistor in series with the battery, then a voltage divider.
#define TEMPPIN      D5  // yellow Onewire data line for temperature probe.
#define MAX_DISTANCE 1400 // Maximum distance we want to ping for (in mm). 
#define ONBOARDLED   D7
#define WAKEPIN      A2 // white wire, to wake from sleep with button.
#define SENSORS      2 // number of DS18B20 temperature sensors

// Variables:
TCPClient client;  // for ThingSpeak

byte   mac[6]; // the MAC address of the Photon
int    battThreshold = 30; // Percent (from Dead to Full) below which a LOW BATTERY! warning is triggered (should be > 2.5V).
double FullChargeVolts = 6.5;  // set for battery being used.
double DeadBattVolts = 5.5;  // Lead Acid batt voltage that the system fails.
float  voltage = 0.0; // Variable to keep track of battery voltage
double BattVolts = 0.0;
double battCalib = 2.16; //correction factor. TO RECALIBRATE THIS, measure batt at its terminals with o-scope.
double soc = 0.0; // Variable to keep track of battery's state-of-charge (SOC)
bool   battalert = LOW; // Variable to keep track of whether battery alert has been triggered
double currentConversion = 2.27;  //To convert A/D counts to mA of current in and out of the battery.
float  BattCurrent = 0.0;
int    negCurCounts; 
int    posCurCounts;
int    negCurOffset = 0;
int    posCurOffset = -27; // to calibrate the current-measuring lines, jumper the 1ohm resistor and make current zero.
int    CurCountsDiff = 0;

int    cm = 0;  // Distance between sensor and oil surface. NewPing returns it in whole cm.
bool   LowOilAlert = LOW; // Goes high when oil level gets below threshold (50mm presently)
int    cm_of_oil = 200; // This is zero when it gets down to the spigot, not the bottom of the tank.
int    mm_of_oil = 2000; //This is zero when it gets down to the spigot, not the bottom of the tank.
int    litres_of_oil = 0;
int    ZeroOilmm = 1304; // When my tank begins to suck air, the mm reading is this.

int    difference = 0; // variable to hold temporary info during calculations
int    q; // for loop counters.
int    SleepSecs = 60; // seconds for device to sleep each loop.
unsigned long lastUpdate = 0;

float  BoxCelsius = NAN;
float  OutsideCelsius = NAN;
String BoxTempStr;
String OutsideTempStr;
DS18B20  TempSensors(TEMPPIN); //Sets name for calls to DS18B20 library, Pin D5 for 1-wire Temp Sensors.
retained uint8_t sensorAddresses[SENSORS][8];  // DS18B20 addresses retained across deep sleep in backup RAM as long as power is on VBAT or VIN of Photon
  
unsigned long ThingSpeakChannelNumber = ; //Enter your Channel number
const char *  ThingSpeakWriteAPIKey = "";  // Enter your API key
  
  // US-100 ultrasonic rangefinder:
 unsigned int  MSByteDistance = 0;
 unsigned int  LSByteDistance = 0;
 unsigned int  mmDistance = 0;
 int           Median_mm = 0;                // Some of these I tried to use unsigned int and got an "ambiguous" compile error.
 int           TempmmDistance = 0;
 int           US100tempData = 0;
 int           junk;
 unsigned long beginmillis = 0;
 unsigned long nowmillis = 0;
 unsigned long elapsedmillis = 0;
 RunningMedian US100reading = RunningMedian(9);
 
  
 // SETUP ______________________________________________________________________________________________

void setup() {
  
  pinMode(ONBOARDLED, OUTPUT);
  pinMode(VBATT, INPUT);
  pinMode(POSCURPIN, INPUT);
  pinMode(NEGCURPIN, INPUT);
  pinMode(WAKEPIN, INPUT_PULLDOWN);
  
  ThingSpeak.begin(client); 
  Particle.variable("Box °C", BoxTempStr);  // temp sensor in box.
  Particle.variable("mm_of_oil",   mm_of_oil); // mm of oil remaining in tank (to spigot).
  Particle.variable("L of Oil",   litres_of_oil); // litres of oil remaining in tank (to spigot).
  Particle.variable("mmDistance",   mm_of_oil); // Average mm measured by rangefinder.
  Particle.variable("BattVolts",   BattVolts); // battery voltage.
  Particle.variable("BattPercent", soc);  // Li-ion battery State of Charge
  Particle.variable("Tank °C", US100tempData);  // Thermometer in ultrasonic sensor
  Particle.variable("Median_mm", Median_mm);  // Median mm measured by rangefinder.
  Particle.variable("Outside °C", OutsideTempStr);  // DS18B20 temperature outside.
  Particle.variable("Raw mm data", TempmmDistance);
  	// To read the values from a browser, go to:
	// http://api.particle.io/v1/devices/{DEVICE_ID}/{VARIABLE}?access_token={ACCESS_TOKEN}

  STARTUP(WiFi.selectAntenna(ANT_EXTERNAL)); // selects the external antenna connector, not chip antenna.
  Serial.begin(115200); // Open serial monitor at 115200 baud to see ping results.
  Serial.println("OilWatcher by Steve M. Potter. Last update 2018-10-03");
  Serial1.begin(9600, SERIAL_8N1); // for comm to/from the US-100 via RX and TX pins.

  WiFi.macAddress(mac);
  Serial.print("MAC address of Photon: ");
  Serial.print(mac[0]&0x0f,HEX);  // This may be in reverse order. 
  Serial.print(":");
  Serial.print(mac[1]&0x0f,HEX);
  Serial.print(":");
  Serial.print(mac[2]&0x0f,HEX);
  Serial.print(":");
  Serial.print(mac[3]&0x0f,HEX);
  Serial.print(":");
  Serial.print(mac[4]&0x0f,HEX);
  Serial.print(":");
  Serial.println(mac[5]&0x0f,HEX);
	
  TempSensors.resetsearch();                 // initialise for DS18B20 temperature sensor search
  for (int i = 0; i < SENSORS; i++) 
   {   
    TempSensors.search(sensorAddresses[i]); //  try to read the 1-wire sensor addresses and if available, store them.
   }
	
} // end setup()

// MAIN LOOP _________________________________________________________________________________________________

void loop() {
  mmDistance  = 0;
  int DataToAvg = 9;
  for (int avgloop = 1; avgloop < (DataToAvg + 1); avgloop++)
  {
   Serial1.flush(); // Clear the serial1 buffer. 
   Serial1.write(0x55); // Send a "distance measure" command to US-100
   delay(200); // US100 response time depends on distance.
   if (Serial1.available() >= 2)  // at least 2 bytes are in buffer
    {
        MSByteDistance = Serial1.read(); // Read both bytes
        LSByteDistance  = Serial1.read();//
        TempmmDistance = (MSByteDistance * 256 + LSByteDistance);
        Particle.publish("Raw mm data", String(TempmmDistance), 60, PRIVATE);
        delay(200);
        mmDistance  = mmDistance + TempmmDistance; // calculate distance in mm. Add to running sum.
        if ((TempmmDistance > 50) && (TempmmDistance < MAX_DISTANCE))   // Test that the distance is in range.
         {
          US100reading.add(TempmmDistance); // put another datum into the buffer for a running median.
         } else    avgloop--; // discard this datum.
    } 
  }
  mmDistance = mmDistance/DataToAvg; // calculate the mean of N measurements.
  Serial.print("Raw mm data: ");
  Serial.print(TempmmDistance);
  Median_mm = US100reading.getMedian(); // get the current running median value of mm from sensor to surface of oil.
     Particle.publish("Median_mm", String(Median_mm), 60, PRIVATE);
  Serial.print("   Median mm data: ");
  Serial.println(Median_mm);
  mm_of_oil = ZeroOilmm - Median_mm; // I determined that my boiler quits when mm = 1304.
  Particle.publish("mm_of_oil", String(mm_of_oil), 60, PRIVATE);
  delay(1000); // wait for last publish to finish.
  if ((mm_of_oil > 0) && (mm_of_oil < MAX_DISTANCE))
  {
   
   ThingSpeak.setField(1, mm_of_oil);
   

   ThingSpeak.setField(7, TempmmDistance);
   delay(1000); // wait for last Publish to finish.
   // Calculate litres....Our tank is 1.10 mm/L 
   litres_of_oil = 0.909 * mm_of_oil;  // Calibrated 2018-02-24 during tank filling.
   Particle.publish("Oil level (L)", String(litres_of_oil), 60, PRIVATE);
   ThingSpeak.setField(6, litres_of_oil);
  }

// Monitor the battery voltage
  voltage = 0;
  for (int z = 1; z < 101; z++)  // It is pretty noisy so we take an average of 100 values.
  {
  voltage += battCalib * ((analogRead(VBATT) / 4096.0) * 3.3); // battery monitor is via a voltage divider.
  }
  voltage = (voltage/100);
  BattVolts = voltage;
  soc = ((voltage - DeadBattVolts) / (FullChargeVolts - DeadBattVolts)) * 100.0;
  Particle.publish("BattVolts",   String(BattVolts, 2), 60, PRIVATE);
  ThingSpeak.setField(2,voltage);
  delay(1000); // Can only publish 4 variables per second. 
  Particle.publish("BattPercent", String(soc, 1), 60, PRIVATE);
 
// Monitor the charge/discharge current. Negative values mean the solar panel is effectively charging the battery.
  CurCountsDiff = 0;
  for (q = 1; q < 101; q++)   // The readings are so noisy, need to average 100 of them.
  {
  negCurCounts = analogRead(NEGCURPIN);
  posCurCounts = (analogRead(POSCURPIN) + posCurOffset);
  difference = (posCurCounts - negCurCounts);
  CurCountsDiff = CurCountsDiff + difference;
  }
  CurCountsDiff = (CurCountsDiff/100);
  
  BattCurrent = (currentConversion * (CurCountsDiff)); 
  Particle.publish("BattCurrent mA", String(BattCurrent, 0), 60, PRIVATE);
  ThingSpeak.setField(4, BattCurrent);
 
 // Send Alerts via Pushbullet:
  if ((mm_of_oil < 50) && (mm_of_oil > 45))  // Only want Pushbullet to send a few warnings
        {
         Particle.publish("Alert", "Heating Oil Level is 5cm - Refill SOON!", PRIVATE); 
         LowOilAlert = HIGH; 
        }
  if (soc < battThreshold) 
        {
	     Particle.publish("Alert", "Oilwatcher battery is very low. Recharge it.", PRIVATE); 
	     battalert = HIGH;
	    }

// Read temperature from the US-100 ultrasonic rangefinder's temp sensor at the top of the tank. The tank air heats up in the sun.
	Serial1.flush(); 
	 while (Serial1.available() >= 1)  // seemed like flush() was not working so I added this.
	 { 
	     junk = Serial1.read();
	 }
    Serial1.write(0x50); // send command to request temperature byte.
    delay(50); // temp response takes about 2ms after command ends.
    if (Serial1.available() >= 1) 
    {
        US100tempData = Serial1.read(); 
        if ((US100tempData > 1) && (US100tempData < 130)) 
        {
            US100tempData -= 45; // Correct by the 45 degree offset of the US100. 
            Particle.publish("Tank °C", String(US100tempData), 60, PRIVATE);
            ThingSpeak.setField(5, US100tempData);
        }
    }

// Read temperature from the DS18B20 sensors in electronics box and outside.
  BoxCelsius = TempSensors.getTemperature(sensorAddresses[1]);  // Will it always find the 1-wire sensor addresses in the same order?
  BoxTempStr = String(BoxCelsius, 1); 
  Particle.publish("Box °C", BoxTempStr, PRIVATE); 
  ThingSpeak.setField(3, BoxTempStr);
  
  OutsideCelsius = TempSensors.getTemperature(sensorAddresses[0]);
  OutsideTempStr = String(OutsideCelsius, 1);
  Particle.publish("Outside °C", OutsideTempStr, PRIVATE); 
  ThingSpeak.setField(8, OutsideTempStr);
  
//  Send all data to ThingSpeak for graphing and calculations.
  ThingSpeak.writeFields(ThingSpeakChannelNumber, ThingSpeakWriteAPIKey); 
                 // give time for temperature to publish before going to sleep.
  delay(10000);  // In case I need to reprogram it, here is a chance when it is not sleeping.
  
  
  System.sleep(WAKEPIN, RISING, SleepSecs);  // Turn off WiFi and pause execution, preserving variables. Wakes after set # of seconds
                                      // A momentary button on the box allows waking for easier re-programming.
                                      // Hit Flash very soon after hitting the wakeup button. Or just before.
  //System.reset();  // This was one way to allow it to not lose its ability connect to wifi when the router's IP address changed during sleep.
                   // Disadvantage is it only loops once thru the code. No variables remembered.
  

}  // end void loop()
